# DATA-DRIVEN KERNEL DESIGNS FOR OPTIMIZED GREEDY SCHEMES: A MACHINE LEARNING PERSPECTIVE by
# TIZIAN WENZEL, FRANCESCO MARCHETTI, EMMA PERRACCHIONE
# Code related to the numerical experiment within Section 4.2. "Machine learning data
# sets", especially to produce Figure 5.


import numpy as np
from matplotlib import pyplot as plt
from scipy import io
import os
import scipy
import matplotlib


np.random.seed(1)


## Some settings
list_datasets = ['fried', 'sarcos', 'ct', 'diamonds', 'stock', 'kegg_undir_uci', 'online_video',
                 'wecs', 'mlr_knn_rng', 'query_agg_count', 'sgemm', 'road_network']

path_for_results = os.path.abspath(os.path.join(os.path.dirname(__file__), 'results_5reruns/')) + '/'


## Initialize dictionaries to store several quantities
# Write all results into a dictionary:
#   dic_eigvals: ratio of cumsummed absolute singular values
#   dic_accuracy_deep: test rmse of 2L model
#   dic_accuracy_beststd: test rmse of best standard model
dic_singular_vals = {}
dic_accuracy_deep = {}
dic_accuracy_beststd = {}
dic_timings = {}
for name_dataset in list_datasets:
    dic_singular_vals[name_dataset] = {}
    dic_accuracy_deep[name_dataset] = {}
    dic_accuracy_beststd[name_dataset] = {}
    dic_timings[name_dataset] = {'list_timings_1L': [], 'list_timings_2L': []}


## Read all file, perform computations, save results
flag_first_npy_file = True
for idx_file, file in enumerate(os.listdir(path_for_results)):
    if file == 'README.md' or file == 'save':
        continue

    name_dataset = '_'.join(file.split('_')[:-1])
    idx_rerun = int(file[-5])

    # Compute and store singular value ratios - enforce that they are ordered!
    results = scipy.io.loadmat(path_for_results + file)
    A_final = results['array_concatenate'][:, :, -1]

    singular_vals = np.linalg.svd(A_final)[1]
    abs_sing_vals = np.abs(singular_vals)
    ordered_abs_sing_vals = np.sort(abs_sing_vals)[::-1]

    ratio = np.cumsum(ordered_abs_sing_vals) / np.sum(ordered_abs_sing_vals)

    dic_singular_vals[name_dataset][idx_rerun] = ratio

    # Compute and store accuracies as well as timings
    dic_accuracy_deep[name_dataset][idx_rerun] = results['array_test_rmse_deep']
    dic_accuracy_beststd[name_dataset][idx_rerun] = np.min(results['array_test_rmse_cv'], axis=0)

    dic_timings[name_dataset]['list_timings_1L'].append(np.array(results['list_timings_1L']))
    dic_timings[name_dataset]['list_timings_2L'].append(np.array(results['list_timings_2L']))


## Calculate ranking where 2L performs best to worst: Use the average mean of improvement for this
array_ratio = np.zeros(len(list_datasets))

for idx_dataset, name_dataset in enumerate(list_datasets):

    array_deep = np.stack([dic_accuracy_deep[name_dataset][idx_rerun] for idx_rerun in range(5)])
    array_beststd = np.stack([dic_accuracy_beststd[name_dataset][idx_rerun] for idx_rerun in range(5)])

    ratio = np.mean(np.mean(array_deep, axis=0) / np.mean(array_beststd, axis=0))

    array_ratio[idx_dataset] = ratio

indices_sorted = np.argsort(array_ratio)


## Print the calculated ratio: This shows, when 2L is superior
print('List indicating when two-layered kernel optimization is beneficial:')
for idx_sorted in indices_sorted:
    print('{:20}'.format(list_datasets[idx_sorted]), np.round(array_ratio[idx_sorted], 5))


## Print timing numbers
for idx_dataset, name_dataset in enumerate(list_datasets):
    array_timings_1L = np.stack(dic_timings[name_dataset]['list_timings_1L'])
    array_timings_2L = np.stack(dic_timings[name_dataset]['list_timings_2L'])

    t_mean_1L = np.mean(array_timings_1L)
    t_std_1L = np.std(array_timings_1L)

    t_mean_2L = np.mean(array_timings_2L, axis=0)
    t_std_2L = np.std(array_timings_2L, axis=0)

    print('1L runtime: {:6.2f} +- {:5.2f}s. '.format(t_mean_1L, t_std_1L)
          + '2L optim: {:6.2f} +- {:5.2f}. '.format(t_mean_2L[0, 0], t_std_2L[0, 0])
          + '2L greedy: {:6.2f} +- {:5.2f}. '.format(t_mean_2L[0, 1], t_std_2L[0, 1])
          + '(' + name_dataset + ')')


## Visualization of the ratio of singular values compared to sum of all singular values
matplotlib.use('TKAgg')

R = np.linspace(0, 1, int(1.5*len(list_datasets)))
array_color=plt.cm.hsv(R)

dic_to_store = {}

list_legend = []
plt.figure(2)
plt.clf()
idx_color = len(indices_sorted)
for idx_sorted in indices_sorted[::-1]:
    name_dataset = list_datasets[idx_sorted]
    array = np.stack([dic_singular_vals[name_dataset][idx_rerun] for idx_rerun in range(5)])
    plt.errorbar(np.linspace(0, 1, array.shape[1]), np.mean(array, axis=0), yerr=np.std(array, axis=0),
                 color=array_color[idx_color, :])
    # list_legend.append('{} '.format(idx_color) + name_dataset)
    list_legend.append(name_dataset)
    idx_color -= 1

    dic_to_store[name_dataset] = array
plt.legend(list_legend)
plt.draw()

## Store in matlab for beautiful tikzfigure plots
# io.savemat('dic_singular_value_ratio.mat', dic_to_store)





