# DATA-DRIVEN KERNEL DESIGNS FOR OPTIMIZED GREEDY SCHEMES: A MACHINE LEARNING PERSPECTIVE by
# TIZIAN WENZEL, FRANCESCO MARCHETTI, EMMA PERRACCHIONE
# Code related to the numerical experiment within Section 4.3. "Stability of the kernel
# optimization", especially to produce the results for Figure 6.


import os
import numpy as np
from scipy.stats import ortho_group  # Requires version 0.18 of scipy
import pickle

from utils_code.dataset_collection import Dataset
from utils_code.hyperparameters import dic_hyperparams
from utils_code.main_function import run_everything


np.random.seed(1)


## Some settings
list_nctrs = [int(np.round(nr)) for nr in np.logspace(np.log(10) / np.log(10), np.log(1000) / np.log(10), 10)]

path_for_results = os.path.abspath(os.path.join(os.path.dirname(__file__), 'results_stability/')) + '/'


flag_gaussian = False
n_reruns = 5

list_datasets = ['fried', 'sarcos', 'ct', 'diamonds', 'stock', 'kegg_undir_uci', 'online_video',
                 'wecs', 'mlr_knn_rng', 'query_agg_count', 'sgemm', 'road_network']


## Loop over all datasets
for idx_dataset, name_dataset in enumerate(list_datasets):

    # Load dataset
    dataset = Dataset()
    X, y = dataset.get_data(name_dataset)
    hyperparameter = dic_hyperparams[name_dataset]

    # list_A_start = [np.eye(X.shape[1])] + [ortho_group.rvs(dim=X.shape[1]) for _ in range(n_reruns - 1)]
    list_A_start = [np.eye(X.shape[1]) for _ in range(n_reruns)]

    # Go only for a fixed set of indices - do not aim at error bars plot!
    list_idx = [0, 1, 2, 3, 4]

    ## Loop to compute and store all initial and final matrices for the different nfolds optimizations
    dic_results = {}
    for idx_index in list_idx:
        dic_results[idx_index] = {}

        list_nfolds = [64, 32, 16, 8, 4]
        assert len(list_nfolds) == len(list_A_start), 'list_nfolds is not of same length as list_A_start'

        ## Load the hyperparameters
        hyperparameter = dic_hyperparams[name_dataset]

        ## Run everything for 2L
        for nfold in list_nfolds:

            A_start, A_optimized, model, model_vkoga1, model_vkoga2, data, _, array_test_rmse_deep, _ = run_everything(
                name_dataset,
                hyperparameter.maxIter_vkoga, hyperparameter.N_points,
                hyperparameter.noise_level, hyperparameter.reg_para_optim, hyperparameter.reg_para_vkoga,
                hyperparameter.learning_rate, hyperparameter.n_epochs,
                hyperparameter.batch_size,
                n_folds=nfold,
                flag_initialize_diagonal=hyperparameter.flag_initialize_diagonal,
                flag_symmetric_A=hyperparameter.flag_symmetric,
                flag_gaussian=hyperparameter.flag_gaussian,
                flag_optim_verbose=False,
                flag_std_vkoga=False,
                idx_rerun=idx_index)

            # Store the initial and the final matrix
            dic_results[idx_index][nfold] = [A_start, A_optimized, model_vkoga1.train_hist['f'],
                                             array_test_rmse_deep]


        ## Run everything for 1L (there is only one!) --> nfold = 0
        A_start, A_optimized, _, model_vkoga1, _, _, _, array_test_rmse_deep, _ = run_everything(
            name_dataset,
            hyperparameter.maxIter_vkoga, hyperparameter.N_points,
            hyperparameter.noise_level, hyperparameter.reg_para_optim, hyperparameter.reg_para_vkoga,
            hyperparameter.learning_rate, n_epochs=0,
            batch_size=hyperparameter.batch_size,
            n_folds=list_nfolds[0], # 64 <-- does not matter as long as we use n_epochs=0
            flag_initialize_diagonal=hyperparameter.flag_initialize_diagonal,
            flag_symmetric_A=hyperparameter.flag_symmetric,
            flag_gaussian=hyperparameter.flag_gaussian,
            flag_optim_verbose=False,
            flag_std_vkoga=False,
            idx_rerun=idx_index)

        # Store the initial and the final matrix
        dic_results[idx_index][0] = [A_start, A_optimized, model_vkoga1.train_hist['f'],
                                     array_test_rmse_deep]


        ## Pickle store dic results
        os.makedirs(path_for_results, exist_ok=True)
        filename = 'dic_results_{}_'.format(idx_index) + name_dataset + '.pkl'
        file_pickle = os.path.join(path_for_results, filename)
        with open(file_pickle, 'wb') as f:
            pickle.dump(dic_results, f)

