# DATA-DRIVEN KERNEL DESIGNS FOR OPTIMIZED GREEDY SCHEMES: A MACHINE LEARNING PERSPECTIVE by
# TIZIAN WENZEL, FRANCESCO MARCHETTI, EMMA PERRACCHIONE
# Code related to the numerical experiment within Section 4.3. "Stability of the kernel
# optimization", especially to produce the plots for Figure 6.


import os
from scipy import io
import numpy as np
from scipy.stats import ortho_group  # Requires version 0.18 of scipy
from matplotlib import pyplot as plt
import pickle
import scipy


np.random.seed(1)


## Some settings
list_nctrs = [int(np.round(nr)) for nr in np.logspace(np.log(10) / np.log(10), np.log(1000) / np.log(10), 10)]

path_for_results = os.path.abspath(os.path.join(os.path.dirname(__file__), 'results_stability/')) + '/'


## First, collect results from all files in a common dictionary
dic_results = {}        # dic_results --> name_dataset --> idx_index --> nfold --> [A_start, A_optimized, model_vkoga1.train_hist['f'], array_test_rmse_deep]
for idx_file, file in enumerate(os.listdir(path_for_results)):
    if 'README' in file:
        continue

    idx_index = int(file.split('_')[2])
    name_dataset = ('_'.join(file.split('_')[3:])).split('.')[:-1][0]
    print(name_dataset)

    # Load file
    with open(path_for_results + file, 'rb') as f:
        dic_test_preds = pickle.load(f)

    # Key name_dataset already available?
    if name_dataset not in dic_results:
        dic_results[name_dataset] = {}

    # Enter all the data
    for idx_index in dic_test_preds:
        dic_results[name_dataset][idx_index] = dic_test_preds[idx_index]



## Second, loop over name_dataset and compute singular value decompositions
dic_singvals = {}               # dic_singvals --> name_dataset --> idx_index --> nfold --> [A_start, A_optimized, model_vkoga1.train_hist['f'], array_test_rmse_deep]
dic_singvecs = {}

for idx_dataset, name_dataset in enumerate(dic_results):

    # if name_dataset != 'wecs' and name_dataset != 'ct' and name_dataset != 'sgemm':
    #     continue

    dic_singvals[name_dataset] = {}
    dic_singvecs[name_dataset] = {}

    for idx_index in dic_results[name_dataset]:

        dic_singvals[name_dataset][idx_index] = {}
        dic_singvecs[name_dataset][idx_index] = {}

        for nfold in dic_results[name_dataset][idx_index]:

            # Compute singular values of finally optimized matrix
            A_start = dic_results[name_dataset][idx_index][nfold][0]
            A_optimized = dic_results[name_dataset][idx_index][nfold][1]

            # Use singular value decomposition
            u, singular_vals, vh = np.linalg.svd(A_optimized)  # A == (u @ np.diag(s)) @ vh
            indices = np.argsort(np.abs(singular_vals))
            indices = indices[::-1]

            singular_vals = singular_vals[indices]
            singular_vecs = u[:, indices]

            dic_singvals[name_dataset][idx_index][nfold] = singular_vals
            dic_singvecs[name_dataset][idx_index][nfold] = singular_vecs



## Third, loop over name_dataset and compute quantities for plots and visualize them (average over the reruns)
max_dim = 15  # maximum dimension to which we want to check similarity

nfold_ref = 64  # we will use this as a reference model!
list_idx = [0, 1, 2, 3, 4]

for idx_dataset, name_dataset in enumerate(list(dic_results.keys())):

    # Skip several datasets
    if name_dataset in ['methane', 'poker', 'protein']:
        continue
    if name_dataset not in ['kegg_undir_uci', 'fried', 'ct', 'mlr_knn_rng']:
        continue

    max_dim1 = min([max_dim, dic_singvals[name_dataset][idx_index][nfold].shape[0]])


    ## Compute quantities for the subsequent plots
    dic_alignment = {}
    dic_singvalsarray = {}

    for idx_nfold, nfold in enumerate(list(dic_results[name_dataset][0].keys())):

        # if nfold == nfold_ref:
        #     continue  # this is our reference model


        # Average results over the different runs (i.e. the different training test splits)
        list_list_similarities = []
        for idx_index in list_idx:

            list_similarity = []
            for dim in range(max_dim1):
                array = scipy.linalg.subspace_angles(dic_singvecs[name_dataset][idx_index][nfold][:, :(dim + 1)],
                                                            dic_singvecs[name_dataset][idx_index][nfold_ref][:, :(dim + 1)])

                # Maybe modify this measure here? 90 degress means orthogonality!!
                similarity = array[0]
                list_similarity.append(np.rad2deg(similarity))

            list_list_similarities.append(list_similarity)

        array_angles = np.stack(list_list_similarities)

        # Write values to dic - to store it: Added "A" in the beginning because otherwise matlab raises an error
        dic_alignment['A{}'.format(nfold)] = array_angles
        dic_singvalsarray['A{}'.format(nfold)] = np.stack([dic_singvals[name_dataset][idx_index][nfold] for idx_index in dic_singvals[name_dataset]])


    ## Errorbar plot of subspace angles
    plt.figure(800 + idx_dataset)
    plt.clf()
    list_legend = []
    for idx_nfold, nfold in enumerate(list(dic_results[name_dataset][0].keys())):
        plt.errorbar(x=np.arange(1, 1 + max_dim1) + (-2 + idx_nfold) * .07,  # shift of 0.07 otherwise errorbars overlap
                     y=np.mean(dic_alignment['A{}'.format(nfold)], axis=0),
                     yerr=np.std(dic_alignment['A{}'.format(nfold)], axis=0))
        list_legend.append('{} vs {}'.format(nfold_ref, nfold if not nfold == 0 else '1L'))

    plt.title(name_dataset + '(dim={}): Alignment of subspaces'.format(dic_singvals[name_dataset][idx_index][nfold].shape[0]))
    plt.legend(list_legend)


    ## Errorbar plot of singular values
    plt.figure(900 + idx_dataset)
    plt.clf()
    list_legend = []
    for idx_nfold, nfold in enumerate(list(dic_results[name_dataset][0].keys())):
        plt.errorbar(x=np.arange(1, 1 + dic_singvalsarray['A0'].shape[1]) + (-2 + idx_nfold) * .07,
                     y=np.mean(dic_singvalsarray['A{}'.format(nfold)], axis=0),
                     yerr=np.std(dic_singvalsarray['A{}'.format(nfold)], axis=0))
        list_legend.append('{} vs {}'.format(nfold_ref, nfold if not nfold == 0 else '1L'))

    plt.title(name_dataset + '(dim={}): Singular values'.format(dic_singvals[name_dataset][idx_index][nfold].shape[0]))
    plt.legend(list_legend)
    plt.yscale('log')


    # ## Store in matlab for beautiful tikzfigure plots
    # io.savemat('dic_stabilityalignment_' + name_dataset + '_nfolds.mat', dic_alignment)
    # io.savemat('dic_singvals_' + name_dataset + '_nfolds.mat', dic_singvalsarray)

    if idx_dataset > 20:
        break


