

from utils_code.optimized_kernel import OptimizedKernel
from utils_code.dataset_collection import Dataset
from utils_code.hyperparameters import dic_hyperparams

from utils_code import tkernels, kernels
import torch
from matplotlib import pyplot as plt
import numpy as np
import time

from utils_code.vkoga import VKOGA
import os

from datetime import datetime


## Some settings
list_nctrs = [int(np.round(nr)) for nr in np.logspace(np.log(10) / np.log(10), np.log(1000) / np.log(10), 10)]

path_for_indices = os.path.abspath(os.path.join(os.path.dirname( __file__ ), '..', 'data/')) + '/'


# Main functiono to run kernel optimization with subsequent vkoga
def run_everything(name_dataset, maxIter_vkoga, N_points, noise_level, reg_para_optim, reg_para_vkoga,
                   learning_rate, n_epochs, batch_size, n_folds,
                   flag_initialize_diagonal=False, flag_symmetric_A=False, flag_gaussian=True,
                   flag_vkoga_verbose=False, flag_plots=False, flag_optim_verbose=True,
                   flag_std_vkoga=True, idx_rerun=None):

    list_timings = []

    ## Load dataset
    dataset = Dataset(N_points=N_points)
    X, y = dataset.get_data(name_dataset)
    hyperparameter = dic_hyperparams[name_dataset]

    # Preprocessing
    if 'example' in name_dataset:
        # No need to shuffle as dataset is randomly generated
        idx = np.arange(X.shape[0])
    else:
        # Random but fixed indices were removed, instead use randomly shuffled every time
        idx = np.arange(X.shape[0])
        np.random.shuffle(idx)

    n_train = int(.8 * X.shape[0])

    X_train = X[idx[:n_train]]
    y_train = y[idx[:n_train]]
    X_test = X[idx[n_train:]]
    y_test = y[idx[n_train:]]
    X_train_torch, y_train_torch = torch.from_numpy(X_train).type(torch.float), torch.from_numpy(y_train).type(torch.float)


    ## Select kernel
    if flag_gaussian:
        kernel_t = tkernels.Gaussian(ep=hyperparameter.shape_para / np.sqrt(X_train.shape[1]))
        kernel = kernels.Gaussian(ep=hyperparameter.shape_para /np.sqrt(X_train.shape[1]))
    else:
        kernel_t = tkernels.Matern(k=hyperparameter.k_matern, ep=hyperparameter.shape_para / np.sqrt(X_train.shape[1]))
        kernel = kernels.Matern(k=hyperparameter.k_matern, ep=hyperparameter.shape_para / np.sqrt(X_train.shape[1]))



    ## Optimize kernel model
    model = OptimizedKernel(kernel=kernel_t, dim=X_train_torch.shape[1],
                            reg_para=reg_para_optim,
                            learning_rate=learning_rate,
                            n_epochs=n_epochs,
                            batch_size=batch_size,
                            n_folds=n_folds,
                            flag_initialize_diagonal=flag_initialize_diagonal,
                            flag_symmetric_A=flag_symmetric_A)

    A_start = torch.clone(model.A).detach().numpy()

    t_optim_start = time.time()
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, '2layered kernel optimization started.')
    model.optimize(X_train_torch, y_train_torch, flag_optim_verbose=flag_optim_verbose)
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, '2layered kernel optimization finished.')
    t_optim_stop = time.time()

    list_timings.append(t_optim_stop - t_optim_start)

    ## Application of VKOGA
    # VKOGA with modified modified kernel, simply pre-apply the linear transformation
    model_vkoga1 = VKOGA(kernel=kernel, greedy_type='f_greedy',
                         verbose=flag_vkoga_verbose, reg_par=reg_para_vkoga)
    t_vkoga1_start = time.time()
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, '2layered VKOGA started.')
    _ = model_vkoga1.fit(X_train @ model.A.detach().numpy(), y_train,
                         X_val=X_test @ model.A.detach().numpy(), y_val=y_test,
                         maxIter=maxIter_vkoga)
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, '2layered VKOGA finished.')
    t_vkoga1_stop = time.time()

    list_timings.append(t_vkoga1_stop - t_vkoga1_start)

    # VKOGA with standard kernel
    model_vkoga2 = VKOGA(kernel=kernel, greedy_type='f_greedy',
                         verbose=flag_vkoga_verbose, reg_par=reg_para_vkoga)
    t_vkoga2_start = time.time()
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Standard VKOGA started.')
    if flag_std_vkoga:
        _ = model_vkoga2.fit(X_train, y_train,
                             X_val=X_test, y_val=y_test,
                             maxIter=maxIter_vkoga)
        print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Standard VKOGA finished.')
    else:
        print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Standard VKOGA was skipped.')

    t_vkoga2_stop = time.time()



    ## Some analysis: Print timings and error
    mse_test_error_vkoga1 = np.linalg.norm(model_vkoga1.predict(X_test @ model.A.detach().numpy()) - y_test) ** 2  / y_test.shape[0]
    if flag_std_vkoga:
        mse_test_error_vkoga2 = np.linalg.norm(model_vkoga2.predict(X_test) - y_test) ** 2 / y_test.shape[0]
    else:
        mse_test_error_vkoga2 = np.linalg.norm(y_test) ** 2 / y_test.shape[0]


    print(name_dataset, 'Deep Kernel Optimization     {:.3f}s.'.format(t_optim_stop - t_optim_start))
    print(name_dataset, 'Deep VKOGA runtime:          {:.3f}s.'.format(t_vkoga1_stop - t_vkoga1_start))
    print(name_dataset, 'Standard VKOGA runtime:      {:.3f}s.'.format(t_vkoga2_stop - t_vkoga2_start))
    print(' -----------------------------------------')
    print(name_dataset, 'Deep kernel test MSE:        {:.3e}'.format(mse_test_error_vkoga1))
    if flag_std_vkoga:
        print(name_dataset, 'Standard kernel test MSE:    {:.3e}'.format(mse_test_error_vkoga2))
    else:
        print(name_dataset, 'Standard kernel test MSE:    skipped.')
    print(' ')

    ## Compute (intermediate) test rmse
    list_nctrs_rmse = list(np.array(list_nctrs)[np.array(list_nctrs) <= maxIter_vkoga])
    array_test_rmse = np.zeros((1, len(list_nctrs_rmse)))

    print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Started computation of test errors.')
    for idx_nctrs, nctrs in enumerate(list_nctrs_rmse): # centers need not to be transformed by matrix A !!
        y_test_pred = kernel.eval(X_test @ model.A.detach().numpy(), model_vkoga1.ctrs_[:nctrs]) \
                      @ model_vkoga1.Cut_[:nctrs, :nctrs].transpose() @ model_vkoga1.c[:nctrs]

        array_test_rmse[0, idx_nctrs] = \
            np.linalg.norm(y_test - y_test_pred) ** 2 / y_test.shape[0]
    print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Finished computation of test errors.')


    ## Some visualizations
    array_concatenate = np.stack(model.list_parameters, 2)
    if flag_plots:
        # Change of parameters of linear layer
        n_epochs = array_concatenate.shape[-1]

        idx_figure_offset = 0
        plt.figure(idx_figure_offset + 5)
        plt.clf()
        list_legend = []
        for k in range(array_concatenate.shape[0]):
            plt.plot(np.arange(1, n_epochs + 1), array_concatenate[k, k, :])
            list_legend.append('A[{},{}]'.format(k, k))
        plt.plot(np.arange(1, n_epochs + 1), array_concatenate[0, 1, :], '--')
        plt.plot(np.arange(1, n_epochs + 1), array_concatenate[1, 0, :], '--')
        plt.legend(list_legend + ['A[0,1]', 'A[1,0]'])
        plt.title('matrix entries')
        plt.xscale('log')
        plt.xlabel('iterations')
        plt.draw()

        # Decay of optimization objective
        plt.figure(idx_figure_offset + 1)
        plt.clf()
        plt.plot(np.arange(1, len(model.list_obj) + 1), model.list_obj)
        plt.xscale('log')
        plt.yscale('log')


        # Decay of vkoga quantities
        plt.figure(idx_figure_offset + 2)
        plt.clf()
        plt.plot(np.arange(1, len(model_vkoga1.train_hist['f']) + 1), model_vkoga1.train_hist['f'], 'm')
        plt.plot(np.arange(1, len(model_vkoga2.train_hist['f']) + 1), model_vkoga2.train_hist['f'], 'b')
        plt.plot(np.arange(1, len(model_vkoga1.train_hist['f val']) + 1), model_vkoga1.train_hist['f val'], 'm--')
        plt.plot(np.arange(1, len(model_vkoga2.train_hist['f val']) + 1), model_vkoga2.train_hist['f val'], 'b--')
        plt.plot(np.arange(1, len(model_vkoga1.train_hist['f val']) + 1), model_vkoga1.train_hist['f val'], 'm--')
        plt.xscale('log')
        plt.yscale('log')
        plt.legend(['2 layer kernel', 'standard'])

        # Plot of predictions
        plt.figure(idx_figure_offset + 3)
        plt.clf()
        plt.plot(y_train, 'k')
        plt.plot(model_vkoga1.predict(X_train @ model.A.detach().numpy()), 'm')
        plt.plot(model_vkoga2.predict(X_train), 'b')
        plt.legend(['true', 'deep kernel', 'standard kernel'])
        plt.title('Train set predictions')

        plt.figure(idx_figure_offset + 4)
        plt.clf()
        plt.plot(y_test, 'k')
        plt.plot(model_vkoga1.predict(X_test @ model.A.detach().numpy()), 'm')
        plt.plot(model_vkoga2.predict(X_test), 'b')
        plt.legend(['true', 'deep kernel', 'standard kernel'])
        plt.title('Test set predictions')

        plt.figure(idx_figure_offset + 40)
        plt.clf()
        # plt.plot(y_test, 'k')
        plt.plot(model_vkoga1.predict(X_test @ model.A.detach().numpy()).reshape(-1) - y_test.reshape(-1), 'm')
        plt.plot(model_vkoga2.predict(X_test).reshape(-1) - y_test.reshape(-1), 'b')
        plt.legend(['deep kernel', 'standard kernel'])
        plt.title('Test set residuals')



    data = [X_train, X_test, y_train, y_test]

    return A_start, model.A.detach().numpy(), model, model_vkoga1, model_vkoga2, \
           data, array_concatenate, array_test_rmse, list_timings



def run_cross_validation(name_dataset, data, maxIter_vkoga, flag_gaussian=True, n_cross_val=5,
                         flag_vkoga_verbose=False):

    # Unpack data, some settings
    [X_train, X_test, y_train, y_test] = data
    hyperparameter = dic_hyperparams[name_dataset]

    lower_bound = .05
    upper_bound = 10

    array_eps = np.logspace(np.log(lower_bound) / np.log(10),
                            np.log(upper_bound) / np.log(10), n_cross_val)
    list_nctrs_rmse = list(np.array(list_nctrs)[np.array(list_nctrs) <= maxIter_vkoga])

    array_test_rmse = np.zeros((n_cross_val, len(list_nctrs_rmse)))

    # Prepare for computation
    list_timings = []
    array_f = np.zeros((maxIter_vkoga, n_cross_val))
    array_f_val = np.zeros((maxIter_vkoga, n_cross_val))

    for idx_para, para_ep in enumerate(array_eps):
        # Select kernel
        if flag_gaussian:
            kernel = kernels.Gaussian(ep=para_ep * 1 / np.sqrt(X_train.shape[1]))
        else:
            kernel = kernels.Matern(k=hyperparameter.k_matern, ep=para_ep * 1 / np.sqrt(X_train.shape[1]))

        # VKOGA with standard kernel
        model_vkoga = VKOGA(kernel=kernel, greedy_type='f_greedy', verbose=flag_vkoga_verbose)
        t_vkoga_start = time.time()
        _ = model_vkoga.fit(X_train, y_train,
                             X_val=X_test, y_val=y_test,
                             maxIter=maxIter_vkoga)
        t_vkoga_stop = time.time()


        array_f[:, idx_para] = model_vkoga.train_hist['f']
        array_f_val[:, idx_para] = model_vkoga.train_hist['f val']


        list_timings.append(t_vkoga_stop - t_vkoga_start)

        print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'kernel_par = {:.2e} run took {:.3f}s.'.format(para_ep, t_vkoga_stop - t_vkoga_start))

        ## Compute (intermediate) test rmse
        print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Started computation of test errors.')
        for idx_nctrs, nctrs in enumerate(list_nctrs_rmse):
            y_test_pred = kernel.eval(X_test, model_vkoga.ctrs_[:nctrs]) \
                          @ model_vkoga.Cut_[:nctrs, :nctrs].transpose() @ model_vkoga.c[:nctrs]

            array_test_rmse[idx_para, idx_nctrs] = \
                np.linalg.norm(y_test - y_test_pred)**2 / y_test.shape[0]
        print(datetime.now().strftime("%H:%M:%S"), name_dataset, 'Finished computation of test errors.')

    return array_eps, array_f, array_f_val, array_test_rmse, list_timings

